@@ -1,6 +1,6 @@
 # Compartmented models base class
 #
-# Copyright (C) 2017--2019 Simon Dobson
+# Copyright (C) 2017--2020 Simon Dobson
 # 
 # This file is part of epydemic, epidemic network simulations in Python.
 #
@@ -17,12 +17,29 @@
 # You should have received a copy of the GNU General Public License
 # along with epydemic. If not, see <http://www.gnu.org/licenses/gpl.html>.
 
-from epydemic import *
+from epydemic import Locus, Process
+import random
 import numpy
 import collections
 
 
-class CompartmentedNodeLocus(Locus):
+class CompartmentedLocus(Locus):
+    '''A locus based on the compartments that nodes reside in.
+
+    :param name: the locus' name'''
+
+    def __init__(self, name):
+        super(CompartmentedLocus, self).__init__(name)
+
+    def compartments(self):
+        '''Return the compartments this locus monitors. This should be
+        overridden by sub-classes as required.
+
+        :returns: the compartments'''
+        return []
+   
+
+class CompartmentedNodeLocus(CompartmentedLocus):
     '''A locus for dynamics occurring at a single node, where the
     locus tracks all nodes in a single compartment.
 
@@ -33,10 +50,49 @@ class CompartmentedNodeLocus(Locus):
         super(CompartmentedNodeLocus, self).__init__(name)
         self._compartment = c
 
+    def compartments(self):
+        '''Return the node compartment we monitor.
 
-class CompartmentedEdgeLocus(Locus):
+        :returns: the compartment'''
+        return [ self._compartment ]
+
+    def addHandler(self, g, n):
+        '''A nodeis added to the network. 
+
+        :param, g: the network
+        :param n: the node'''
+        if not isinstance(n, tuple):
+            super(CompartmentedNodeLocus, self).addHandler(g, n)
+
+    def leaveHandler(self, g, n):
+        '''A node changes compartment. 
+
+        :param, g: the network
+        :param n: the node'''
+        if not isinstance(n, tuple):
+            super(CompartmentedNodeLocus, self).leaveHandler(g, n)
+
+    def enterHandler(self, g, n):
+        '''A node enters a compartment. 
+
+        :param, g: the network
+        :param n: the node'''
+        if not isinstance(n, tuple):
+            super(CompartmentedNodeLocus, self).enterHandler(g, n)
+
+    def removeHandler(self, g, n):
+        '''A node is removed from the network. 
+
+        :param, g: the network
+        :param n: the node'''
+        if not isinstance(n, tuple):
+            super(CompartmentedNodeLocus, self).removeHandler(g, n)
+
+
+class CompartmentedEdgeLocus(CompartmentedLocus):
     '''A locus for dynamics occurring at an edge, where the locus tracks all
-    edges whose endpoint nodes are in specified compartments.
+    edges whose endpoint nodes are in specified compartments. Since the network may
+    also be adaptive, we need to track additions and removals of edges too.
 
     :param name: the locus' name
     :param l: the left compartment
@@ -47,6 +103,12 @@ class CompartmentedEdgeLocus(Locus):
         self._left = l
         self._right = r
 
+    def compartments(self):
+        '''Return the compartments of the node endpoints we monitor.
+
+        :returns: the compartments'''
+        return [ self._left, self._right ]
+
     def matches(self, g, n, m):
         '''Test whether the given edge has the right compartment endpoints for this compartment. The
         method returns 1 if the edge has the right compartments in the orientation (n, m), -1 if it has
@@ -64,6 +126,24 @@ class CompartmentedEdgeLocus(Locus):
             else:
                 return 0
 
+    def addHandler(self, g, e):
+        '''An edge is added to the network, check if its endpoint compartments match the
+        locus and add it if so. 
+
+        :param, g: the network
+        :param e: the edge'''
+        if isinstance(e, tuple):
+            (n, m) = e
+            es = self.elements()
+            match = self.matches(g, n, m)
+            if match == -1:
+                #print('edge ({m}, {n}) added to {l}'.format(n = n, m = m, l = self._name))
+                es.add((m, n))
+            else:
+                if match == 1:
+                    #print('edge ({n}, {m}) added {l}'.format(n = n, m = m, l = self._name))
+                    es.add((n, m))
+
     def leaveHandler(self, g, n):
         '''Node leaves one of the edge's compartments, remove any incident edges
         that no longer have the correct orientation.
@@ -74,12 +154,12 @@ class CompartmentedEdgeLocus(Locus):
         for (nn, mm) in g.edges(n):
             match = self.matches(g, nn, mm)
             if match == -1:
-                #print 'edge ({m}, {n}) leaves {l}'.format(n = nn, m = mm, l = self._name)
-                es.remove((mm, nn))
+                #print('edge ({m}, {n}) leaves {l}'.format(n = nn, m = mm, l = self._name))
+                es.discard((mm, nn))
             else:
                 if match == 1:
-                    #print 'edge ({n}, {m}) leaves {l}'.format(n = nn, m = mm, l = self._name)
-                    es.remove((nn, mm))
+                    #print('edge ({n}, {m}) leaves {l}'.format(n = nn, m = mm, l = self._name))
+                    es.discard((nn, mm))
 
     def enterHandler(self, g, n):
         '''Node enters one of the edge's compartments, add any incident edges
@@ -91,13 +171,31 @@ class CompartmentedEdgeLocus(Locus):
         for (nn, mm) in g.edges(n):
             match = self.matches(g, nn, mm)
             if match == -1:
-                # print 'edge ({m}, {n}) enters {l}'.format(n = nn, m = mm, l = self._name)
+                #print('edge ({m}, {n}) enters {l}'.format(n = nn, m = mm, l = self._name))
                 es.add((mm, nn))
             else:
                 if match == 1:
-                    # print 'edge ({n}, {m}) enters {l}'.format(n = nn, m = mm, l = self._name)
+                    #print('edge ({n}, {m}) enters {l}'.format(n = nn, m = mm, l = self._name))
                     es.add((nn, mm))
 
+    def removeHandler(self, g, e):
+        '''An edge is removed from the network, check whether it was in this locus and
+        remove it if so.
+
+        :param, g: the network
+        :param e: the edge'''
+        if isinstance(e, tuple):
+            (n, m) = e
+            es = self.elements()
+            match = self.matches(g, n, m)
+            if match == -1:
+                #print('edge ({m}, {n}) removed from {l}'.format(n = n, m = m, l = self._name))
+                es.discard((m, n))
+            else:
+                if match == 1:
+                    #print('edge ({n}, {m}) removed from {l}'.format(n = n, m = m, l = self._name))
+                    es.discard((n, m))
+
 
 class CompartmentedModel(Process):
     '''The base class a :term:`compartmented model of disease`, which
@@ -116,9 +214,6 @@ class CompartmentedModel(Process):
     OCCUPIED = 'occupied'                       #: Edge attribute, True if infection travelled along the edge.
     T_OCCUPIED = 'occupationTime'               #: Edge attribute holding the time the infection crossed the edge.
 
-    # default compartment
-    DEFAULT_COMPARTMENT = 'defaultCompartment'  #: Default compartment label for new nodes
-
     def __init__( self ):
         super(CompartmentedModel, self).__init__()
 
@@ -128,9 +223,8 @@ class CompartmentedModel(Process):
     def reset( self ):
         '''Reset the model ready to be built.'''
         super(CompartmentedModel, self).reset()
-        self._compartments = []
-        self._effects = dict()
-        self._eventDistribution = None
+        self._compartments = dict()            # compartment -> initial probability
+        self._effects = dict()                 # compartment -> event handlers
 
     def setUp(self, params):
         '''Set up the initial population of nodes into compartments.
@@ -158,7 +252,10 @@ class CompartmentedModel(Process):
         compartment of each node.
 
         :returns: a list of (compartment, probability) pairs'''
-        return self._compartments
+        dist = []
+        for cp in self._compartments.items():
+            dist.append(cp)
+        return dist
 
     def initialCompartments( self ):
         '''Place each node in the network into its initial compartment. The default
@@ -174,7 +271,7 @@ class CompartmentedModel(Process):
         g = self.network()
         for n in g.nodes():
             # select a compartment according to the initial distribution
-            r = numpy.random.random()
+            r = random.random()
             a = 0.0
             for (c, p) in dist:
                 a = a + p
@@ -188,14 +285,20 @@ class CompartmentedModel(Process):
 
     # ---------- Termination and results ----------
 
+    def compartments(self):
+        '''Return the set of compartments.
+
+        :returns: the compartments'''
+        return self._compartments.keys()
+
     def compartment( self, c ):
         '''Return all the nodes currently in a particular compartment in a network. This works
-        for all compartments, not just those that are loci for dynamics.
+        for all compartments, not just those that are loci for dynamics -- but is a *lot*
+        slower, so it's better to create a :term:`locus` is you're going to access a compartment frequently.
 
         :param c: the compartment
         :returns: a collection of nodes'''
-        g = self.network()
-        return [ n for n in g.nodes() if g.nodes[n][self.COMPARTMENT] == c]
+        return [ n for n in self.network().nodes() if self.getCompartment(n) == c ]
 
     def results(self):
         '''Create a dict of experimental results for the experiment, consisting of the final
@@ -204,20 +307,18 @@ class CompartmentedModel(Process):
         :returns: a dict of experimental results'''
         rc = super(CompartmentedModel, self).results()
 
-        for (c, _) in self._compartments:
-            rc[c] = 0
-        g = self.network()
-        for n in g.nodes():
-            c = g.nodes[n][self.COMPARTMENT]
-            rc[c] = rc[c] + 1
+        # add size of each compartment
+        for c in self.compartments():
+            rc[c] = len(self.compartment(c))
         return rc
 
     def skeletonise(self):
         '''Remove unoccupied edges from the network. This leaves the network
         consisting of only "occupied" edges that were used to transmit the
-        infection between nodes. Note that this process means that further
-        dynamics over the network probably don't make sense, unless you're
-        actually wanting to run on the residual network post-infection.
+        infection between nodes, also known as the :term:`contact tree`. Note
+        that this process means that further dynamics over the network probably
+        don't make sense, unless you're actually wanting to run on the residual
+        network post-infection.
 
         :returns: the network with unoccupied edges removed'''
 
@@ -245,82 +346,104 @@ class CompartmentedModel(Process):
 
         :param c: the compartment name
         :param p: the initial occupancy probability (defaults to  0.0)'''
-        self._compartments.append((c, p))
+        self._compartments[c] = p
 
     def trackNodesInCompartment(self, c, name = None):
         '''Add a locus tracking nodes in a given compartment.
 
         :param c: the compartment to track
-        :param name: (optional) the name of the locus (defaults to the compartment name)'''
+        :param name: (optional) the name of the locus (defaults to the compartment name)
+        :returns: the locus used to track the nodes'''
         if name is None:
             name = c
-        if name in self:
-            raise NameError('Locus {c} already exists in model'.format(c = name))
-        else:
-            # add locus
-            locus = CompartmentedNodeLocus(name, c)
-            self.addLocus(name, locus)
 
-            # add handlers
-            self._addHandlers(c, lambda g, n: locus.leaveHandler(g, n),
-                                 lambda g, n: locus.enterHandler(g, n))
+        # add locus
+        locus = CompartmentedNodeLocus(name, c)
+        return self.addLocus(name, locus)
 
     def trackEdgesBetweenCompartments(self, l, r, name = None):
         '''Add a locus to track edges with endpoint nodes in the given compartments.
 
         :param l: the compartment of the left node
         :param r: the compartment of the right node
-        :param name: (optional) the name of the locus (defaults to a combination of the two compartment names)'''
+        :param name: (optional) the name of the locus (defaults to a combination of the two compartment names)
+        :returns: the locus used to track the nodes'''
+ 
         if name is None:
             name = '{l}-{r}'.format(l = l, r = r)
-        if name in self:
-            raise NameError('Locus {c} already exists in model'.format(c = name))
+
+        # add locus
+        locus = CompartmentedEdgeLocus(name, l, r)
+        return self.addLocus(name, locus)
+
+    def addLocus(self, n, l = None):
+        '''Add a locus to the model, initialising the handler functions.
+
+        :param n: the name
+        :param l: (optional) the locus
+        :returns: the locus'''
+        locus = super(CompartmentedModel,self).addLocus(n, l)
+        for c in locus.compartments():
+            if c not in self._effects.keys():
+                self._effects[c] = []
+            self._effects[c].append((locus.addHandler, locus.leaveHandler, locus.enterHandler, locus.removeHandler))
+
+    def _handlerCompartments(self, e):
+        '''Return the compartments that a given element's change might affect.
+
+        :param e: a node or edge
+        :returns: a list of compartments'''
+        g = self.network()
+        if isinstance(e, tuple):
+            # element is an edge, check compartments at the endpoints
+            (n, m) = e
+            cs = [ g.nodes[n][self.COMPARTMENT], g.nodes[m][self.COMPARTMENT] ]
         else:
-            # add locus
-            locus = CompartmentedEdgeLocus(name, l, r)
-            self.addLocus(name, locus)
+            # element is a node, check its own compartmnent
+            cs = [ g.nodes[e][self.COMPARTMENT] ]
+        return cs
 
-            # add handlers
-            self._addHandlers(l, lambda g, n: locus.leaveHandler(g, n),
-                                 lambda g, n: locus.enterHandler(g, n))
-            self._addHandlers(r, lambda g, n: locus.leaveHandler(g, n),
-                                 lambda g, n: locus.enterHandler(g, n))
+    def _callAddHandlers( self, e ):
+        '''Call all handlers affected by a node or edge being added to the network.
 
-    def _addHandlers( self, c, lh, eh ):
-        '''Add handler functions for a compartment, to be called whenever the contents
-        of that compartment change. This links compartments to loci.
+        :param e the node or edge'''
+        g = self.network()
+        for c in self._handlerCompartments(e):
+            if c in self._effects.keys():
+                for (ah, _, _, _) in self._effects[c]:
+                    ah(g, e)
 
-        Handlers are functions that take a model, a network, a node, and a
-        compartment, and provide the appropriate behaviour for when the node
-        enters (or leaves) the compartment.
-        
-        :param c: the compartment
-        :param lh: the leave handler
-        :param eh: the enter handler'''
-        if c not in self._effects.keys():
-            self._effects[c] = [ (lh, eh) ]
-        else:
-            self._effects[c].append((lh, eh))
+    def _callLeaveHandlers( self, e, c ):
+        '''Call all handlers affected by a node or edge leaving a compartment.
 
-    def _callLeaveHandlers( self, n, c ):
-        '''Call all handlers affected by a node leaving a compartment.
-
-        :param n: the node
+        :param e: the node or edge
         :param c: the compartment'''
-        if c in self._effects.keys():
-            g = self.network()
-            for (lh, _) in self._effects[c]:
-                lh(g, n)
+        g = self.network()
+        for c in self._handlerCompartments(e):
+            if c in self._effects.keys():
+                for (_, lh, _, _) in self._effects[c]:
+                    lh(g, e)
 
-    def _callEnterHandlers( self, n, c ):
-        '''Call all handlers affected by a node entering a compartment.
+    def _callEnterHandlers( self, e, c ):
+        '''Call all handlers affected by a node or edge entering a compartment.
 
-        :param n: the node
+        :param e: the node or edge
         :param c: the compartment'''
-        if c in self._effects.keys():
-            g = self.network()
-            for (_, eh) in self._effects[c]:
-                eh(g, n)
+        g = self.network()
+        for c in self._handlerCompartments(e):
+            if c in self._effects.keys():
+                for (_, _, eh, _) in self._effects[c]:
+                    eh(g, e) 
+
+    def _callRemoveHandlers( self, e ):
+        '''Call all handlers affected by a node or edge being removed from the network.
+
+        :param e: the node or edge'''
+        g = self.network()
+        for c in self._handlerCompartments(e):
+            if c in self._effects.keys():
+                for (_, _, _, rh) in self._effects[c]:
+                    rh(g, e)
 
 
     # ---------- Accessing and evolving the network ----------
@@ -340,15 +463,24 @@ class CompartmentedModel(Process):
         # propagate the change to any other compartments
         self._callEnterHandlers(n, c)
 
+    def getCompartment(self, n):
+        '''Return the compartment of a node.
+
+        :parak n: the node
+        :returns: its compartment'''
+        return self.network().nodes[n][self.COMPARTMENT]
+
     def changeCompartment( self, n, c ):
         '''Change the compartment of a node.
 
         :param n: the node
         :param c: the new compartment for the node'''
         g = self.network()
+        oc = g.nodes[n][self.COMPARTMENT]
 
         # propagate effects of leaving the current compartment
-        self._callLeaveHandlers(n, g.nodes[n][self.COMPARTMENT])
+        if oc is not None:
+            self._callLeaveHandlers(n, oc)
 
         # record new compartment on node
         g.nodes[n][self.COMPARTMENT] = c
@@ -368,25 +500,22 @@ class CompartmentedModel(Process):
         data[self.OCCUPIED] = True
         data[self.T_OCCUPIED] = t
         
-    def addNode(self, n, compartment = None, **kwds):
+    def addNode(self, n, c, **kwds):
         '''Add a node to the working network, adding it to the appropriate compartment.
 
         :param n: the new node
-        :param compartment: (optional) compartment for the node (defaults to :attr:`DEFAULT_COMPARTMENT`)
+        :param c: compartment for the node
         :param kwds: (optional) node attributes'''
         super(CompartmentedModel, self).addNode(n, **kwds)
-        if compartment is None:
-            compartment = self.DEFAULT_COMPARTMENT
-        self.setCompartment(n, compartment)
+        self.setCompartment(n, c)
 
     def removeNode(self, n):
         '''Remove a node from the working network, updating any affected compartments.
 
         :param n: the node'''
 
-        # propagate effects of leaving the current compartment
-        g = self.network()
-        self._callLeaveHandlers(n, g.nodes[n][self.COMPARTMENT])
+        # remove node from any loci, and from its compartment
+        self._callRemoveHandlers(n)
 
         # remove the node
         super(CompartmentedModel, self).removeNode(n)
@@ -400,19 +529,16 @@ class CompartmentedModel(Process):
         super(CompartmentedModel, self).addEdge(n, m, **kwds)
 
         # add edge to any compartments it should be in
-        g = self.network()
-        self._callEnterHandlers(n, g.nodes[n][self.COMPARTMENT])
-        self._callEnterHandlers(m, g.nodes[m][self.COMPARTMENT])
-
+        self._callAddHandlers((n, m))
 
     def removeEdge(self, n, m):
         '''Remove an edge from the working network and from any compartments.
 
         :param n: the start node
         :param m: the end node'''
-        super(CompartmentedModel, self).removeEdge(n, m)
 
-        # add edge to any compartments it should be in
-        g = self.network()
-        self._callLeaveandlers(n, g.nodes[n][self.COMPARTMENT])
-        self._callLeaveHandlers(m, g.nodes[m][self.COMPARTMENT])
+        # remove edge from any compartments it should be in
+        self._callRemoveHandlers((n, m))
+
+        # remove the edge from the network
+        super(CompartmentedModel, self).removeEdge(n, m)
